/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the Open-Xchange, Inc. group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2004-2010 Open-Xchange, Inc.
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
package com.openexchange.usm.connector.commands;

import static com.openexchange.usm.connector.commands.CommandConstants.*;

import java.util.BitSet;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.openexchange.usm.api.contenttypes.ContentType;
import com.openexchange.usm.api.contenttypes.ContentTypeField;
import com.openexchange.usm.api.contenttypes.DefaultContentTypes;
import com.openexchange.usm.api.contenttypes.FolderConstants;
import com.openexchange.usm.api.contenttypes.FolderContentType;
import com.openexchange.usm.api.contenttypes.FolderContentTypeTransferHandler;
import com.openexchange.usm.api.database.DatabaseAccessException;
import com.openexchange.usm.api.datatypes.DataType;
import com.openexchange.usm.api.datatypes.PIMAttachment;
import com.openexchange.usm.api.datatypes.PIMAttachments;
import com.openexchange.usm.api.exceptions.USMAccessDeniedException;
import com.openexchange.usm.api.exceptions.USMException;
import com.openexchange.usm.api.exceptions.USMSQLException;
import com.openexchange.usm.api.session.ChangeState;
import com.openexchange.usm.api.session.DataObject;
import com.openexchange.usm.api.session.Folder;
import com.openexchange.usm.api.session.FolderType;
import com.openexchange.usm.api.session.Session;
import com.openexchange.usm.datatypes.contacts.ContactDistributionMember;
import com.openexchange.usm.datatypes.contacts.Image;
import com.openexchange.usm.datatypes.tasks.calendar.CommonCalendarTasksFieldNames;
import com.openexchange.usm.json.ConnectorBundleErrorCodes;
import com.openexchange.usm.json.JSONSessionInitializer;
import com.openexchange.usm.json.USMJSONAPIException;
import com.openexchange.usm.json.USMJSONServlet;
import com.openexchange.usm.json.USMSessionID;
import com.openexchange.usm.json.response.ResponseStatusCode;
import com.openexchange.usm.session.dataobject.DataObjectUtil;
import com.openexchange.usm.util.Toolkit;

public abstract class NormalCommandHandler extends BaseCommandHandler {

	private static final String DELETE_EXCEPTIONS = "delete_exceptions";

	protected DataObject[] _allFolders;

	protected NormalCommandHandler(USMJSONServlet servlet, HttpSession httpSession, JSONObject parameters,
			HttpServletRequest request) throws USMJSONAPIException {
		super(servlet, httpSession, parameters, request);
		String sessionID = getStringParameter(SESSIONID);
		USMSessionID usmSessionID = (USMSessionID) httpSession.getAttribute(sessionID);
		if (usmSessionID == null)
			throw new USMJSONAPIException(ConnectorBundleErrorCodes.COMMAND_UNKNOWN_SESSION,
					ResponseStatusCode.UNKNOWN_SESSION, "Unknown " + SESSIONID);
		try {
			setSession(servlet.getSession(usmSessionID.getUser(), usmSessionID.getPassword(), usmSessionID.getDevice(),
					getClientIP(), getXHeadersFromCurrentRequest()));
		} catch (USMAccessDeniedException e) {
			throw new USMJSONAPIException(ConnectorBundleErrorCodes.COMMAND_ACCESS_DENIED,
					ResponseStatusCode.ACCESS_DENIED, e.getErrorMessage());
		} catch (USMException e) {
			throw USMJSONAPIException.createInternalError(ConnectorBundleErrorCodes.COMMAND_INTERNAL_ERROR, e);
		}
	}

	protected FolderContentType getFolderContentType() {
		return (FolderContentType) _servlet.getContentTypeManager().getContentType(DefaultContentTypes.FOLDER_ID);
	}

	protected void storeExtraFields(ContentType type, BitSet extraFields) throws DatabaseAccessException,
			USMSQLException {
		JSONSessionInitializer.storeExtraFields(_session, type, extraFields);
	}

	protected JSONObject storeCompleteDataObjectInJSONObject(DataObject o) throws JSONException, USMJSONAPIException {
		return storeCompleteDataObjectInJSONObject(o, null);
	}

	protected JSONObject storeCompleteDataObjectInJSONObject(DataObject o,
			Map<DataObject, List<DataObject>> exceptionsMap) throws JSONException, USMJSONAPIException {
		List<DataObject> exceptions = getExceptionsOfDataObject(o, exceptionsMap);
		JSONObject newObject = storeDataObjectFieldsInJSONObject(o, exceptions);
		if (exceptions != null)
			storeExceptionsInJSONObject(newObject, exceptions);
		newObject.put(OBJECT_TYPE, o.getContentType().getID());
		newObject.put(UUID_KEY, o.getUUID().toString());
		return newObject;
	}

	private List<DataObject> getExceptionsOfDataObject(DataObject o, Map<DataObject, List<DataObject>> exceptionsMap) {
		if (exceptionsMap != null) {
			for (Map.Entry<DataObject, List<DataObject>> entry : exceptionsMap.entrySet()) {
				DataObject object = entry.getKey();
				if (object.getID().equals(o.getID()))
					return entry.getValue();
			}
		}
		return null;
	}

	private void storeExceptionsInJSONObject(JSONObject jo, List<DataObject> exceptions) throws USMJSONAPIException,
			JSONException {
		JSONArray exceptionsJSONArray = new JSONArray();

		JSONArray deleteExceptions = jo.optJSONArray(CommonCalendarTasksFieldNames.DELETE_EXCEPTIONS);
		for (DataObject changeException : exceptions) {
			boolean existsInDeleteExceptions = false;
			for (int i = 0; deleteExceptions != null && i < deleteExceptions.length(); i++) {
				if (changeException.getFieldContent(CommonCalendarTasksFieldNames.RECURRENCE_DATE_POSITION).equals(
						deleteExceptions.get(i)))
					existsInDeleteExceptions = true;
			}
			if (!existsInDeleteExceptions) {
				JSONObject exceptionJSONObject = storeCompleteDataObjectInJSONObject(changeException);
				exceptionsJSONArray.put(exceptionJSONObject);
			}
		}
		jo.put(CHANGE_EXCEPTIONS, exceptionsJSONArray);
	}

	protected JSONObject storeDataObjectFieldsInJSONObject(DataObject o, List<DataObject> exceptions)
			throws JSONException, USMJSONAPIException {
		JSONObject newObject = new JSONObject();
		ContentTypeField[] fields = o.getContentType().getFields();
		for (int i = 0; i < fields.length; i++) {
			DataType<?> fieldType = fields[i].getFieldType();
			Object v = o.getFieldContent(i);
			String fieldName = fields[i].getFieldName();
			if (shouldIncludeFieldInClientResult(o, i, fieldType, v, fieldName)) {
				if (DefaultContentTypes.MAIL_ID.equals(o.getContentType().getID())
						&& JSONSessionInitializer.isInternalMailField(fieldName))
					continue;
				if (IMAGE1.equals(fieldType.getName())) {
					if (v == null) {
						newObject.put(IMAGE1, JSONObject.NULL);
					} else {
						byte[] data = ((Image) v).getData();
						newObject.put(IMAGE1, (data == null) ? JSONObject.NULL : Toolkit.encodeBase64(data));
					}
				} else if (ATTACHMENTS_LAST_MODIFIED.equals(fieldName)) {
					storeAttachmentsInJSONObject(o.getSession(), newObject, v,
							convertToPIMAttachmentDataType(fieldType), o);
				} else if (NUMBER_OF_ATTACHMENTS.equals(fieldName)) {
					fieldType.storeInJSONObject(o.getSession(), newObject, fieldName, v);
					if (!o.isFieldModified(ATTACHMENTS_LAST_MODIFIED)) {
						for (int j = 0; j < fields.length; j++) {
							if (ATTACHMENTS_LAST_MODIFIED.equals(fields[j].getFieldName())) {
								fieldType = fields[j].getFieldType();
								v = o.getFieldContent(j);
								storeAttachmentsInJSONObject(o.getSession(), newObject, v,
										convertToPIMAttachmentDataType(fieldType), o);
							}
						}
					}
				} else if (CommandConstants.DISTRIBUTION_LIST.equals(fieldName)) {
					fieldType.storeInJSONObject(o.getSession(), newObject, fieldName, v);
					JSONArray list = newObject.getJSONArray(CommandConstants.DISTRIBUTION_LIST);
					JSONArray newList = new JSONArray();
					int length = list.length();
					for (int j = 0; j < length; j++) {
						ContactDistributionMember dMember = new ContactDistributionMember(list.getJSONObject(j));
						JSONObject jObj = dMember.toJSONObject();
						String user_id = dMember.getUser_id();
						if (user_id != null) {
							try {
								UUID uuid = o.getSession().getUUID(o.getContentType(), Integer.parseInt(user_id));
								if (uuid != null)
									jObj.put(UUID_KEY, uuid.toString());
							} catch (NumberFormatException ignored) {
								//do nothing
							} catch (USMException ignored) {
								//do nothing
							}
						}
						newList.put(jObj);
					}
					newObject.put(CommandConstants.DISTRIBUTION_LIST, newList);
				} else if (o instanceof Folder && TITLE.equals(fieldName)) {
					String folderTitleForClient = generateFolderTitleForClient((Folder) o, v);
					fieldType.storeInJSONObject(o.getSession(), newObject, fieldName, folderTitleForClient);
				} else {
					fieldType.storeInJSONObject(o.getSession(), newObject, fieldName, v);
				}
			}
		}

		if (o.getContentType().getCode() == DefaultContentTypes.FOLDER_CODE) {
			addOptionalFolderUUID(o, newObject);
		} else if (o.getContentType().getCode() == DefaultContentTypes.CALENDAR_CODE && !isSimpleAppointment(o)
				&& !isAppException(o)) {
			JSONArray deleteExceptions = new JSONArray();
			boolean foundChange = false;
			Object v = o.getFieldContent(DELETE_EXCEPTIONS);
			if (v instanceof Object[]) {
				for (Object e : ((Object[]) v))
					deleteExceptions.put(e);
				v = o.getFieldContent(CHANGE_EXCEPTIONS);
				if (v instanceof Object[]) {
					for (Object e : ((Object[]) v)) {
						if (isChangeExceptionOutOfSyncRange(e)
								|| (!isParentFolderShared(o) && !isCurrentUserParticipantToChangeException(exceptions,
										e)))
							deleteExceptions.put(e);
						else
							foundChange = true;
					}
				}
			}
			// We need to always report delete_exceptions because change_exceptions that are outside of the
			// sync interval have to be reported as delete exceptions. This means that we can not rely on the
			// modification state of the property because even if the property hasn't changed we will report
			// a different content because a change_exception moved into or out of the sync interval.
			// Only for new creations and an empty delete_exceptions list we may omit the field.
			if (deleteExceptions.length() > 0 || o.getChangeState() != ChangeState.CREATED) {
				newObject.put(DELETE_EXCEPTIONS, deleteExceptions);
			}

			if (!foundChange) {
				if (o.getChangeState() == ChangeState.MODIFIED && o.isFieldModified(CHANGE_EXCEPTIONS)) {
					v = o.getFieldContent(CHANGE_EXCEPTIONS);
					if (!(v instanceof Object[]) || ((Object[]) v).length == 0)
						newObject.put(CHANGE_EXCEPTIONS, new JSONArray());
				}
			}
		}
		return newObject;
	}

	private boolean isParentFolderShared(DataObject o) {
		Folder parentFolder = o.getParentFolder();
		if (parentFolder == null)
			try {
				parentFolder = _session.getCachedFolder(o.getParentFolderID());
			} catch (DatabaseAccessException e) {
				return false;
			} catch (USMSQLException e) {
				return false;
			}
		if (parentFolder == null)
			return false;
		else
			return FolderType.SHARED.matches(parentFolder.getFieldContent(FolderConstants.TYPE));
	}

	private boolean isCurrentUserParticipantToChangeException(List<DataObject> exceptions, Object e) {
		DataObject exceptionObject = findDataObjectByRecurrenceDatePosition(exceptions, e);
		int currentUser = _session.getUserIdentifier();
		if (exceptionObject != null) {
			int attendeeStatus = SyncCommandHandler.getAttendeeStatus(exceptionObject, currentUser);
			return attendeeStatus != -1 && attendeeStatus != 2;
		}
		return false;
	}

	private DataObject findDataObjectByRecurrenceDatePosition(Collection<DataObject> list, Object position) {
		if (list != null && position != null) {
			for (DataObject o : list) {
				if (position.equals(o.getFieldContent(CommonCalendarTasksFieldNames.RECURRENCE_DATE_POSITION)))
					return o;
			}
		}
		return null;
	}

	private boolean isChangeExceptionOutOfSyncRange(Object e) {
		return ((Number) e).longValue() < _session.getStartDate() || ((Number) e).longValue() > _session.getEndDate();
	}

	protected boolean shouldIncludeFieldInClientResult(DataObject o, int i, DataType<?> fieldType, Object v,
			String fieldName) {
		return (o.getChangeState() != ChangeState.MODIFIED) ? !fieldType.isDefaultValue(v)
				&& !shouldFilterOutField(fieldName) : o.isFieldModified(i) && isFieldSynchronized(o, i)
				&& !shouldFilterOutField(fieldName);
	}

	protected boolean shouldFilterOutField(String fieldName) {
		return CHANGE_EXCEPTIONS.equals(fieldName) || RECURRENCE_ID.equals(fieldName);
	}

	protected boolean isFieldSynchronized(DataObject o, int index) {
		BitSet contentTypeFilter = _session.getFieldFilter(o.getContentType());
		return contentTypeFilter.get(index);
	}

	protected boolean isSimpleAppointment(DataObject dataObject) {
		if (dataObject.getContentType().getCode() != DefaultContentTypes.CALENDAR_CODE)
			return false;
		Object o = dataObject.getFieldContent(CommonCalendarTasksFieldNames.RECURRENCE_TYPE);
		return (o instanceof Number) && ((Number) o).intValue() == 0;
	}

	protected boolean isAppException(DataObject dataObject) {
		if (dataObject.getContentType().getCode() != DefaultContentTypes.CALENDAR_CODE)
			return false;
		String id = dataObject.getID();
		Number recurrence_id = (dataObject.getFieldContent(RECURRENCE_ID) == null) ? null : (Number) (dataObject
				.getFieldContent(RECURRENCE_ID));
		return recurrence_id != null && recurrence_id.intValue() != 0 && !id.equals(recurrence_id.toString());
	}

	protected void addOptionalFolderUUID(DataObject o, JSONObject object) throws USMJSONAPIException, JSONException {
		if (object.has(FOLDER_ID)) {
			initAllFolders();
			DataObject parent = DataObjectUtil.findDataObject(o.getParentFolderID(), _allFolders);
			object.put(FOLDER_UUID, (parent != null) ? parent.getUUID().toString() : "");
		}
	}

	private void initAllFolders() throws USMJSONAPIException {
		if (_allFolders != null)
			return;
		try {
			_allFolders = _session.getCachedFolders();
		} catch (DatabaseAccessException e) {
			throw generateCantDetermineFolderIDException();
		} catch (USMSQLException e) {
			throw generateCantDetermineFolderIDException();
		}
	}

	@SuppressWarnings("unchecked")
	private DataType<PIMAttachments> convertToPIMAttachmentDataType(DataType<?> fieldType) {
		return (DataType<PIMAttachments>) fieldType;
	}

	private USMJSONAPIException generateCantDetermineFolderIDException() throws USMJSONAPIException {
		return new USMJSONAPIException(ConnectorBundleErrorCodes.COMMAND_CANNOT_DETERMINE_FOLDER_UUID,
				ResponseStatusCode.INTERNAL_ERROR, "Can not determine folder_uuid, list of folders not available");
	}

	private void storeAttachmentsInJSONObject(Session session, JSONObject jObj, Object value,
			DataType<PIMAttachments> fieldType, DataObject dataObject) throws JSONException {
		PIMAttachments attachments = fieldType.checkValue(value);
		JSONArray array = new JSONArray();
		if (attachments == null) //  || attachments.size() == 0
			return;
		for (PIMAttachment att : attachments.getAttachments()) {
			if (att.getData() != null) {
				JSONObject object = att.toJSONObject();
				object.put(UUID_KEY, att.getUUID().toString());
				object.put(CommandConstants.DATA_KEY, Toolkit.encodeBase64(att.getData()));
				array.put(object);
			}
		}
		if (array.length() >= 0)
			jObj.put(ATTACHMENTS, array);
	}

	/**
	 * Bug 17240: when the folder belongs to another user, then that user's name must be used as a prefix for the folder title.
	 * @throws USMJSONAPIException
	 */
	private String generateFolderTitleForClient(Folder folder, Object oxFolderTitleUntyped) throws USMJSONAPIException {
		String oxFolderTitle = (oxFolderTitleUntyped == null) ? "" : oxFolderTitleUntyped.toString();

		String module = folder.getElementsContentTypeID();
		if (!FolderType.SHARED.matches(folder.getFieldContent(TYPE))
				|| FolderContentTypeTransferHandler.MODULE_SYSTEM.equals(module))
			return oxFolderTitle;

		Folder parent = getParentFolder(folder);
		while (parent != null) {
			if (FolderType.SYSTEM.matches(parent.getFieldContent(TYPE))) {
				String folderOwnerName = parent.getFieldContent(TITLE).toString();
				// Bug 18160: prevent double prefixing
				if (oxFolderTitle.startsWith(folderOwnerName))
					return oxFolderTitle;
				return folderOwnerName + " - " + oxFolderTitle;
			}
			parent = getParentFolder(parent);
		}
		_servlet.getJournal().debug(folder.getSession() + " No user name found for shared folder " + folder.getID());
		return oxFolderTitle;
	}

	private Folder getParentFolder(Folder folder) throws USMJSONAPIException {
		Folder parent = folder.getParentFolder();
		if (parent != null)
			return parent;
		initAllFolders();
		DataObject parentUntyped = DataObjectUtil.findDataObject(folder.getParentFolderID(), _allFolders);
		if (parentUntyped instanceof Folder)
			return (Folder) parentUntyped;
		return null;
	}

	protected void logSyncData(String command, String uuid, long oldSyncID, long newSyncID, long tempSyncID) {
		if (uuid == null)
			uuid = "<hierarchy>";
		if (tempSyncID == 0L || tempSyncID == newSyncID)
			_servlet.getJournal().debug(
					_session + command + " modified SyncID(" + uuid + "):" + oldSyncID + "->" + newSyncID);
		else
			_servlet.getJournal().debug(
					_session + command + " modified SyncID(" + uuid + "):" + oldSyncID + "->" + newSyncID + "("
							+ tempSyncID + ")");
	}
}
